@@ -20,7 +20,7 @@
 #include <stdint.h>
 #include <sys/types.h>
 
-#if defined(WIN32)
+#if defined(_WIN32)
 #define EXTERN __declspec( dllexport )
 #else
 #define EXTERN
@@ -30,15 +30,18 @@
 extern "C" {
 #endif
 
+
 struct iscsi_context;
 struct sockaddr;
+struct scsi_iovec;
 
 /* API VERSION */
-#define LIBISCSI_API_VERSION (20150621)
+#define LIBISCSI_API_VERSION (20170105)
 
 /* FEATURES */
 #define LIBISCSI_FEATURE_IOVECTOR (1)
 #define LIBISCSI_FEATURE_NOP_COUNTER (1)
+#define LIBISCSI_FEATURE_ISER (1)
 
 #define MAX_STRING_SIZE (255)
 
@@ -50,6 +53,10 @@ struct sockaddr;
 #define ISCSI_PORTAL_URL_SYNTAX "\"iscsi://[<username>[%<password>]@]" \
   "<host>[:<port>]\""
 
+enum iscsi_transport_type {
+	TCP_TRANSPORT = 0,
+	ISER_TRANSPORT = 1
+};
 
 EXTERN void iscsi_set_cache_allocations(struct iscsi_context *iscsi, int ca);
 
@@ -82,6 +89,11 @@ EXTERN int iscsi_service(struct iscsi_context *iscsi, int revents);
  * How many commands are in flight.
  */
 EXTERN int iscsi_queue_length(struct iscsi_context *iscsi);
+/*
+ * How many commands are queued for dispatch.
+ */
+EXTERN int iscsi_out_queue_length(struct iscsi_context *iscsi);
+
 
 /************************************************************
  * Timeout Handling.
@@ -140,6 +152,7 @@ struct iscsi_url {
        char target_passwd[MAX_STRING_SIZE + 1];
        int lun;
        struct iscsi_context *iscsi;
+       enum iscsi_transport_type transport;
 };
 
 /*
@@ -153,7 +166,7 @@ enum iscsi_immediate_data {
 	ISCSI_IMMEDIATE_DATA_NO  = 0,
 	ISCSI_IMMEDIATE_DATA_YES = 1
 };
-int iscsi_set_immediate_data(struct iscsi_context *iscsi, enum iscsi_immediate_data immediate_data);
+EXTERN int iscsi_set_immediate_data(struct iscsi_context *iscsi, enum iscsi_immediate_data immediate_data);
 
 /*
  * This function is used to set the desired mode for initial_r2t
@@ -166,7 +179,7 @@ enum iscsi_initial_r2t {
 	ISCSI_INITIAL_R2T_NO  = 0,
 	ISCSI_INITIAL_R2T_YES = 1
 };
-int
+EXTERN int
 iscsi_set_initial_r2t(struct iscsi_context *iscsi, enum iscsi_initial_r2t initial_r2t);
 
 
@@ -180,7 +193,7 @@ iscsi_set_initial_r2t(struct iscsi_context *iscsi, enum iscsi_initial_r2t initia
  * "iqn.ronnie.test%3A1234" will be translated to "iqn.ronnie.test:1234"
  *
  * Function will return a pointer to an iscsi url structure if successful,
- * or it will return NULL and set iscsi_get_error() accrodingly if there was a problem
+ * or it will return NULL and set iscsi_get_error() accordingly if there was a problem
  * with the URL.
  *
  * CHAP username/password can also be provided via environment variables
@@ -201,7 +214,7 @@ EXTERN void iscsi_destroy_url(struct iscsi_url *iscsi_url);
  * lun is not yet known.
  *
  * Function will return a pointer to an iscsi url structure if successful,
- * or it will return NULL and set iscsi_get_error() accrodingly if there was a problem
+ * or it will return NULL and set iscsi_get_error() accordingly if there was a problem
  * with the URL.
  *
  * CHAP username/password can also be provided via environment variables
@@ -223,15 +236,15 @@ EXTERN const char *iscsi_get_error(struct iscsi_context *iscsi);
  * Initiator_name is the iqn name we want to identify to the target as.
  *
  * Returns:
- *  0: success
- * <0: error
+ *  non-NULL: success
+ *  NULL: error
  */
 EXTERN struct iscsi_context *iscsi_create_context(const char *initiator_name);
 
 /*
  * Destroy an existing ISCSI context and tear down any existing connection.
  * Callbacks for any command in flight will be invoked with
- * ISCSI_STATUS_CANCELLED.
+ * SCSI_STATUS_CANCELLED.
  *
  * Returns:
  *  0: success
@@ -239,6 +252,18 @@ EXTERN struct iscsi_context *iscsi_create_context(const char *initiator_name);
  */
 EXTERN int iscsi_destroy_context(struct iscsi_context *iscsi);
 
+/*
+ * Sets and initializes the transport type for a context.
+ * TCP_TRANSPORT is the default and is available on all platforms.
+ * ISER_TRANSPORT is conditionally supported on Linux where available.
+ *
+ * Returns:
+ *  0: success
+ * <0: error
+ */
+EXTERN int iscsi_init_transport(struct iscsi_context *iscsi,
+                                enum iscsi_transport_type transport);
+
 /*
  * Set an optional alias name to identify with when connecting to the target
  *
@@ -248,6 +273,7 @@ EXTERN int iscsi_destroy_context(struct iscsi_context *iscsi);
  */
 EXTERN int iscsi_set_alias(struct iscsi_context *iscsi, const char *alias);
 
+
 /*
  * Set the iqn name of the taqget to login to.
  * The target name must be set before a normal-login can be initiated.
@@ -369,13 +395,13 @@ typedef void (*iscsi_command_cb)(struct iscsi_context *iscsi, int status,
  *
  * Callback parameters :
  * status can be either of :
- *    ISCSI_STATUS_GOOD     : Connection was successful. Command_data is NULL.
- *                            In this case the callback will be invoked a
- *                            second time once the connection is torn down.
+ *    SCSI_STATUS_GOOD     : Connection was successful. Command_data is NULL.
+ *                           In this case the callback will be invoked a
+ *                           second time once the connection is torn down.
  *
- *    ISCSI_STATUS_ERROR    : Either failed to establish the connection, or
- *                            an already established connection has failed
- *                            with an error.
+ *    SCSI_STATUS_ERROR    : Either failed to establish the connection, or
+ *                           an already established connection has failed
+ *                           with an error.
  *
  * The callback will NOT be invoked if the session is explicitely torn down
  * through a call to iscsi_disconnect() or iscsi_destroy_context().
@@ -408,13 +434,13 @@ EXTERN int iscsi_connect_sync(struct iscsi_context *iscsi, const char *portal);
  *
  * Callback parameters :
  * status can be either of :
- *    ISCSI_STATUS_GOOD     : Connection was successful. Command_data is NULL.
- *                            In this case the callback will be invoked a
- *                            second time once the connection is torn down.
+ *    SCSI_STATUS_GOOD     : Connection was successful. Command_data is NULL.
+ *                           In this case the callback will be invoked a
+ *                           second time once the connection is torn down.
  *
- *    ISCSI_STATUS_ERROR    : Either failed to establish the connection, or
- *                            an already established connection has failed
- *                            with an error.
+ *    SCSI_STATUS_ERROR    : Either failed to establish the connection, or
+ *                           an already established connection has failed
+ *                           with an error.
  *
  * The callback will NOT be invoked if the session is explicitely torn down
  * through a call to iscsi_disconnect() or iscsi_destroy_context().
@@ -476,10 +502,10 @@ EXTERN int iscsi_reconnect_sync(struct iscsi_context *iscsi);
  *
  * Callback parameters :
  * status can be either of :
- *    ISCSI_STATUS_GOOD     : login was successful. Command_data is always
- *                            NULL.
- *    ISCSI_STATUS_CANCELLED: login was aborted. Command_data is NULL.
- *    ISCSI_STATUS_ERROR    : login failed. Command_data is NULL.
+ *    SCSI_STATUS_GOOD     : login was successful. Command_data is always
+ *                           NULL.
+ *    SCSI_STATUS_CANCELLED: login was aborted. Command_data is NULL.
+ *    SCSI_STATUS_ERROR    : login failed. Command_data is NULL.
  */
 EXTERN int iscsi_login_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 		      void *private_data);
@@ -504,9 +530,9 @@ EXTERN int iscsi_login_sync(struct iscsi_context *iscsi);
  *
  * Callback parameters :
  * status can be either of :
- *    ISCSI_STATUS_GOOD     : logout was successful. Command_data is always
- *                            NULL.
- *    ISCSI_STATUS_CANCELLED: logout was aborted. Command_data is NULL.
+ *    SCSI_STATUS_GOOD     : logout was successful. Command_data is always
+ *                           NULL.
+ *    SCSI_STATUS_CANCELLED: logout was aborted. Command_data is NULL.
  */
 EXTERN int iscsi_logout_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 		       void *private_data);
@@ -521,29 +547,6 @@ EXTERN int iscsi_logout_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 EXTERN int iscsi_logout_sync(struct iscsi_context *iscsi);
 
 
-/*
- * Asynchronous call to perform an ISCSI discovery.
- *
- * discoveries can only be done on connected and logged in discovery sessions.
- *
- * Returns:
- *  0 if the call was initiated and a discovery  will be attempted. Result
- *    of the logout will be reported through the callback function.
- * <0 if there was an error. The callback function will not be invoked.
- *
- * Callback parameters :
- * status can be either of :
- *    ISCSI_STATUS_GOOD     : Discovery was successful. Command_data is a
- *                            pointer to a iscsi_discovery_address list of
- *                            structures.
- *                            This list of structures is only valid for the
- *                            duration of the callback and all data will be
- *                            freed once the callback returns.
- *    ISCSI_STATUS_CANCELLED: Discovery was aborted. Command_data is NULL.
- */
-EXTERN int iscsi_discovery_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
-			  void *private_data);
-
 struct iscsi_target_portal {
        struct iscsi_target_portal *next;
        const char *portal;
@@ -555,6 +558,47 @@ struct iscsi_discovery_address {
        struct iscsi_target_portal *portals;
 };
 
+/*
+ * Asynchronous call to perform an ISCSI discovery.
+ *
+ * discoveries can only be done on connected and logged in discovery sessions.
+ *
+ * Returns:
+ *  0 if the call was initiated and a discovery  will be attempted. Result
+ *    will be reported through the callback function.
+ * <0 if there was an error. The callback function will not be invoked.
+ *
+ * Callback parameters :
+ * status can be either of :
+ *    SCSI_STATUS_GOOD     : Discovery was successful. Command_data is a
+ *                           pointer to a iscsi_discovery_address list of
+ *                           structures.
+ *                           This list of structures is only valid for the
+ *                           duration of the callback and all data will be
+ *                           freed once the callback returns.
+ *    SCSI_STATUS_CANCELLED : Discovery was aborted. Command_data is NULL.
+ */
+EXTERN int iscsi_discovery_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
+			  void *private_data);
+
+/*
+ * Synchronous call to perform an ISCSI discovery.
+ *
+ * discoveries can only be done on connected and logged in discovery sessions.
+ *
+ * Returns:
+ *  NULL if there was an error.
+ *  struct iscsi_discovery_address* if the discovery was successfull.
+ *    The data returned must be released by calling iscsi_free_discovery_data.
+ */
+EXTERN struct iscsi_discovery_address *iscsi_discovery_sync(
+        struct iscsi_context *iscsi);
+
+/* Free the discovery data structures returned by iscsi_discovery_sync
+ */
+EXTERN void iscsi_free_discovery_data(struct iscsi_context *iscsi,
+                                      struct iscsi_discovery_address *da);
+
 /*
  * Asynchronous call to perform an ISCSI NOP-OUT call
  *
@@ -565,11 +609,11 @@ struct iscsi_discovery_address {
  *
  * Callback parameters :
  * status can be either of :
- *    ISCSI_STATUS_GOOD     : NOP-OUT was successful and the server responded
- *                            with a NOP-IN callback_data is a iscsi_data
- *                            structure containing the data returned from
- *                            the server.
- *    ISCSI_STATUS_CANCELLED: Discovery was aborted. Command_data is NULL.
+ *    SCSI_STATUS_GOOD     : NOP-OUT was successful and the server responded
+ *                           with a NOP-IN callback_data is a iscsi_data
+ *                           structure containing the data returned from
+ *                           the server.
+ *    SCSI_STATUS_CANCELLED : Discovery was aborted. Command_data is NULL.
  * 
  * The callback may be NULL if you only want to let libiscsi count the in-flight
  * NOPs.
@@ -595,20 +639,31 @@ enum iscsi_task_mgmt_funcs {
      ISCSI_TM_TASK_REASSIGN     = 0x08
 };
 
+enum iscsi_task_mgmt_response {
+	ISCSI_TMR_FUNC_COMPLETE				= 0x0,
+	ISCSI_TMR_TASK_DOES_NOT_EXIST			= 0x1,
+	ISCSI_TMR_LUN_DOES_NOT_EXIST			= 0x2,
+	ISCSI_TMR_TASK_STILL_ALLEGIANT			= 0x3,
+	ISCSI_TMR_TASK_ALLEGIANCE_REASS_NOT_SUPPORTED	= 0x4,
+	ISCSI_TMR_TMF_NOT_SUPPORTED			= 0x5,
+	ISCSI_TMR_FUNC_AUTH_FAILED			= 0x6,
+	ISCSI_TMR_FUNC_REJECTED				= 0xFF
+};
+
 /*
  * Asynchronous call for task management
  *
  * Returns:
  *  0 if the call was initiated and the task mgmt function will be invoked.
- * the connection will be reported through the callback function.
+ *    The result will be reported through the callback function.
  * <0 if there was an error. The callback function will not be invoked.
  *
  * Callback parameters :
  * status can be either of :
- *    ISCSI_STATUS_GOOD     : Connection was successful. Command_data is a pointer to uint32_t
- *                            containing the response code as per RFC3720/10.6.1
+ *    SCSI_STATUS_GOOD     : Connection was successful. Command_data is a pointer to uint32_t
+ *                           containing the response code as per RFC3720/10.6.1
  *
- *    ISCSI_STATUS_ERROR    : Error.
+ *    SCSI_STATUS_ERROR     : Error.
  *
  * The callback will NOT be invoked if the session is explicitely torn down
  * through a call to iscsi_disconnect() or iscsi_destroy_context().
@@ -679,19 +734,19 @@ iscsi_task_mgmt_target_cold_reset_sync(struct iscsi_context *iscsi);
  * The content of command_data depends on the status type.
  *
  * status :
- *   ISCSI_STATUS_GOOD the scsi command completed successfullt on the target.
+ *   SCSI_STATUS_GOOD the scsi command completed successfullt on the target.
  *   If this scsi command returns DATA-IN, that data is stored in an scsi_task
  *   structure returned in the command_data parameter. This buffer will be
  *   automatically freed once the callback returns.
  *
- *   ISCSI_STATUS_CHECK_CONDITION the scsi command failed with a scsi sense.
+ *   SCSI_STATUS_CHECK_CONDITION the scsi command failed with a scsi sense.
  *   Command_data contains a struct scsi_task. When the callback returns,
  *   this buffer will automatically become freed.
  *
- *   ISCSI_STATUS_CANCELLED the scsi command was aborted. Command_data is
+ *   SCSI_STATUS_CANCELLED the scsi command was aborted. Command_data is
  *   NULL.
  *
- *   ISCSI_STATUS_ERROR the command failed. Command_data is NULL.
+ *   SCSI_STATUS_ERROR the command failed. Command_data is NULL.
  */
 
 struct iscsi_data {
@@ -762,6 +817,18 @@ iscsi_readcapacity16_task(struct iscsi_context *iscsi, int lun,
 			  iscsi_command_cb cb,
 			  void *private_data);
 EXTERN struct scsi_task *
+iscsi_readdefectdata10_task(struct iscsi_context *iscsi, int lun,
+                            int req_plist, int req_glist,
+                            int defect_list_format, uint16_t alloc_len,
+                            iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_readdefectdata12_task(struct iscsi_context *iscsi, int lun,
+                            int req_plist, int req_glist,
+                            int defect_list_format,
+                            uint32_t address_descriptor_index,
+                            uint32_t alloc_len,
+                            iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
 iscsi_sanitize_task(struct iscsi_context *iscsi, int lun,
 		    int immed, int ause, int sa, int param_len,
 		    struct iscsi_data *data,
@@ -807,53 +874,112 @@ EXTERN struct scsi_task *
 iscsi_read6_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		       uint32_t datalen, int blocksize, iscsi_command_cb cb,
 		       void *private_data);
-
+EXTERN struct scsi_task *
+iscsi_read6_iov_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		       uint32_t datalen, int blocksize, iscsi_command_cb cb,
+		       void *private_data, struct scsi_iovec *iov, int niov);
 EXTERN struct scsi_task *
 iscsi_read10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		  uint32_t datalen, int blocksize,
 		  int rdprotect, int dpo, int fua, int fua_nv, int group_number,
 		  iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_read10_iov_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		  uint32_t datalen, int blocksize,
+		  int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		  iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_write10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_write10_iov_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_writeverify10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int bytchk, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_writeverify10_iov_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int bytchk, int group_number,
+		   iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_read12_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   uint32_t datalen, int blocksize,
 		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_read12_iov_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   uint32_t datalen, int blocksize,
+		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_write12_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_write12_iov_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_writeverify12_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int bytchk, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_writeverify12_iov_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int bytchk, int group_number,
+		   iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_read16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   uint32_t datalen, int blocksize,
 		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_read16_iov_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   uint32_t datalen, int blocksize,
+		   int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_write16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_write16_iov_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		   iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
+iscsi_writeatomic16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			 unsigned char *data, uint32_t datalen, int blocksize,
+			 int wrprotect, int dpo, int fua, int group_number,
+		   iscsi_command_cb cb, void *private_data);
+EXTERN struct scsi_task *
+iscsi_writeatomic16_iov_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			     unsigned char *data, uint32_t datalen, int blocksize,
+			     int wrprotect, int dpo, int fua, int group_number,
+			     iscsi_command_cb cb, void *private_data,
+			     struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_orwrite_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_orwrite_iov_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		       unsigned char *data, uint32_t datalen, int blocksize,
+		       int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		       iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_startstopunit_task(struct iscsi_context *iscsi, int lun,
 			 int immed, int pcm, int pc,
 			 int no_flush, int loej, int start,
@@ -868,41 +994,83 @@ iscsi_compareandwrite_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_compareandwrite_iov_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			       unsigned char *data, uint32_t datalen, int blocksize,
+			       int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+			       iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_writeverify16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int bytchk, int group_number,
 		   iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_writeverify16_iov_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		   unsigned char *data, uint32_t datalen, int blocksize,
+		   int wrprotect, int dpo, int bytchk, int group_number,
+		   iscsi_command_cb cb, void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_verify10_task(struct iscsi_context *iscsi, int lun,
 		    unsigned char *data, uint32_t datalen, uint32_t lba,
 		    int vprotect, int dpo, int bytchk,
 		    int blocksize, iscsi_command_cb cb,
 		    void *private_data);
 EXTERN struct scsi_task *
+iscsi_verify10_iov_task(struct iscsi_context *iscsi, int lun,
+			unsigned char *data, uint32_t datalen, uint32_t lba,
+			int vprotect, int dpo, int bytchk,
+			int blocksize, iscsi_command_cb cb,
+			void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_verify12_task(struct iscsi_context *iscsi, int lun,
 		    unsigned char *data, uint32_t datalen, uint32_t lba,
 		    int vprotect, int dpo, int bytchk,
 		    int blocksize, iscsi_command_cb cb,
 		    void *private_data);
 EXTERN struct scsi_task *
+iscsi_verify12_iov_task(struct iscsi_context *iscsi, int lun,
+			unsigned char *data, uint32_t datalen, uint32_t lba,
+			int vprotect, int dpo, int bytchk,
+			int blocksize, iscsi_command_cb cb,
+			void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_verify16_task(struct iscsi_context *iscsi, int lun,
 		    unsigned char *data, uint32_t datalen, uint64_t lba,
 		    int vprotect, int dpo, int bytchk,
 		    int blocksize, iscsi_command_cb cb,
 		    void *private_data);
 EXTERN struct scsi_task *
+iscsi_verify16_iov_task(struct iscsi_context *iscsi, int lun,
+			unsigned char *data, uint32_t datalen, uint64_t lba,
+			int vprotect, int dpo, int bytchk,
+			int blocksize, iscsi_command_cb cb,
+			void *private_data, struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_writesame10_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		       unsigned char *data, uint32_t datalen,
 		       uint16_t num_blocks,
 		       int anchor, int unmap, int wrprotect, int group,
 		       iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_writesame10_iov_task(struct iscsi_context *iscsi, int lun, uint32_t lba,
+			   unsigned char *data, uint32_t datalen,
+			   uint16_t num_blocks,
+			   int anchor, int unmap, int wrprotect, int group,
+			   iscsi_command_cb cb, void *private_data,
+			   struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_writesame16_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		       unsigned char *data, uint32_t datalen,
 		       uint32_t num_blocks,
 		       int anchor, int unmap, int wrprotect, int group,
 		       iscsi_command_cb cb, void *private_data);
 EXTERN struct scsi_task *
+iscsi_writesame16_iov_task(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			   unsigned char *data, uint32_t datalen,
+			   uint32_t num_blocks,
+			   int anchor, int unmap, int wrprotect, int group,
+			   iscsi_command_cb cb, void *private_data,
+			   struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_modeselect6_task(struct iscsi_context *iscsi, int lun,
 		       int pf, int sp, struct scsi_mode_page *mp,
 		       iscsi_command_cb cb, void *private_data);
@@ -961,6 +1129,16 @@ iscsi_report_supported_opcodes_task(struct iscsi_context *iscsi, int lun,
 				    uint32_t alloc_len,
 				    iscsi_command_cb cb, void *private_data);
 
+EXTERN struct scsi_task *
+iscsi_receive_copy_results_task(struct iscsi_context *iscsi, int lun,
+				int sa, int list_id, int alloc_len,
+				iscsi_command_cb cb, void *private_data);
+
+EXTERN struct scsi_task *
+iscsi_extended_copy_task(struct iscsi_context *iscsi, int lun,
+			 struct iscsi_data *param_data,
+			 iscsi_command_cb cb, void *private_data);
+
 /*
  * Sync commands for SCSI
  */
@@ -1001,51 +1179,120 @@ EXTERN struct scsi_task *
 iscsi_read6_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		  uint32_t datalen, int blocksize);
 
+EXTERN struct scsi_task *
+iscsi_read6_iov_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		     uint32_t datalen, int blocksize,
+		     struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_read10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		  uint32_t datalen, int blocksize,
 		  int rdprotect, int dpo, int fua, int fua_nv, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_read10_iov_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		      uint32_t datalen, int blocksize,
+		      int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		      struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_write10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
 
 EXTERN struct scsi_task *
+iscsi_write10_iov_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+			unsigned char *data, uint32_t datalen, int blocksize,
+			int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+			struct scsi_iovec *iov, int niov);
+EXTERN struct scsi_task *
 iscsi_writeverify10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int bytchk, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_writeverify10_iov_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+			     unsigned char *data, uint32_t datalen, int blocksize,
+			     int wrprotect, int dpo, int bytchk, int group_number,
+			     struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_read12_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		  uint32_t datalen, int blocksize,
 		  int rdprotect, int dpo, int fua, int fua_nv, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_read12_iov_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		      uint32_t datalen, int blocksize,
+		      int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		      struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_write12_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_write12_iov_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+		       unsigned char *data, uint32_t datalen, int blocksize,
+		       int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		       struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_writeverify12_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int bytchk, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_writeverify12_iov_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+			     unsigned char *data, uint32_t datalen, int blocksize,
+			     int wrprotect, int dpo, int bytchk, int group_number,
+			     struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_read16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		  uint32_t datalen, int blocksize,
 		  int rdprotect, int dpo, int fua, int fua_nv, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_read16_iov_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		      uint32_t datalen, int blocksize,
+		      int rdprotect, int dpo, int fua, int fua_nv, int group_number,
+		      struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_write16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_write16_iov_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		       unsigned char *data, uint32_t datalen, int blocksize,
+		       int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		       struct scsi_iovec *iov, int niov);
+
+EXTERN struct scsi_task *
+iscsi_writeatomic16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			 unsigned char *data, uint32_t datalen, int blocksize,
+			 int wrprotect, int dpo, int fua, int group_number);
+
+EXTERN struct scsi_task *
+iscsi_writeatomic16_iov_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			     unsigned char *data, uint32_t datalen, int blocksize,
+			     int wrprotect, int dpo, int fua, int group_number,
+			     struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_orwrite_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_orwrite_iov_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+		       unsigned char *data, uint32_t datalen, int blocksize,
+		       int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+		       struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_startstopunit_sync(struct iscsi_context *iscsi, int lun,
 			 int immed, int pcm, int pc,
@@ -1060,11 +1307,23 @@ iscsi_compareandwrite_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int fua, int fua_nv, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_compareandwrite_iov_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			       unsigned char *data, uint32_t datalen, int blocksize,
+			       int wrprotect, int dpo, int fua, int fua_nv, int group_number,
+			       struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_writeverify16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		   unsigned char *data, uint32_t datalen, int blocksize,
 		   int wrprotect, int dpo, int bytchk, int group_number);
 
+EXTERN struct scsi_task *
+iscsi_writeverify16_iov_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			     unsigned char *data, uint32_t datalen, int blocksize,
+			     int wrprotect, int dpo, int bytchk, int group_number,
+			     struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_readcapacity10_sync(struct iscsi_context *iscsi, int lun, int lba,
 			  int pmi);
@@ -1072,6 +1331,17 @@ iscsi_readcapacity10_sync(struct iscsi_context *iscsi, int lun, int lba,
 EXTERN struct scsi_task *
 iscsi_readcapacity16_sync(struct iscsi_context *iscsi, int lun);
 
+EXTERN struct scsi_task *
+iscsi_readdefectdata10_sync(struct iscsi_context *iscsi, int lun,
+                            int req_plist, int req_glist,
+                            int defect_list_format, uint16_t alloc_len);
+
+EXTERN struct scsi_task *
+iscsi_readdefectdata12_sync(struct iscsi_context *iscsi, int lun,
+                            int req_plist, int req_glist,
+                            int defect_list_format,
+                            uint32_t address_descriptor_index,
+                            uint32_t alloc_len);
 EXTERN struct scsi_task *
 iscsi_get_lba_status_sync(struct iscsi_context *iscsi, int lun, uint64_t starting_lba, uint32_t alloc_len);
 
@@ -1110,30 +1380,62 @@ iscsi_verify10_sync(struct iscsi_context *iscsi, int lun,
 		    int vprotect, int dpo, int bytchk,
 		    int blocksize);
 
+EXTERN struct scsi_task *
+iscsi_verify10_iov_sync(struct iscsi_context *iscsi, int lun,
+			unsigned char *data, uint32_t datalen, uint32_t lba,
+			int vprotect, int dpo, int bytchk,
+			int blocksize, struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_verify12_sync(struct iscsi_context *iscsi, int lun,
 		    unsigned char *data, uint32_t datalen, uint32_t lba,
 		    int vprotect, int dpo, int bytchk,
 		    int blocksize);
 
+EXTERN struct scsi_task *
+iscsi_verify12_iov_sync(struct iscsi_context *iscsi, int lun,
+			unsigned char *data, uint32_t datalen, uint32_t lba,
+			int vprotect, int dpo, int bytchk,
+			int blocksize, struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_verify16_sync(struct iscsi_context *iscsi, int lun,
 		    unsigned char *data, uint32_t datalen, uint64_t lba,
 		    int vprotect, int dpo, int bytchk,
 		    int blocksize);
 
+EXTERN struct scsi_task *
+iscsi_verify16_iov_sync(struct iscsi_context *iscsi, int lun,
+			unsigned char *data, uint32_t datalen, uint64_t lba,
+			int vprotect, int dpo, int bytchk,
+			int blocksize, struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_writesame10_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
 		       unsigned char *data, uint32_t datalen,
 		       uint16_t num_blocks,
 		       int anchor, int unmap, int wrprotect, int group);
 
+EXTERN struct scsi_task *
+iscsi_writesame10_iov_sync(struct iscsi_context *iscsi, int lun, uint32_t lba,
+			   unsigned char *data, uint32_t datalen,
+			   uint16_t num_blocks,
+			   int anchor, int unmap, int wrprotect, int group,
+			   struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_writesame16_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
 		       unsigned char *data, uint32_t datalen,
 		       uint32_t num_blocks,
 		       int anchor, int unmap, int wrprotect, int group);
 
+EXTERN struct scsi_task *
+iscsi_writesame16_iov_sync(struct iscsi_context *iscsi, int lun, uint64_t lba,
+			   unsigned char *data, uint32_t datalen,
+			   uint32_t num_blocks,
+			   int anchor, int unmap, int wrprotect, int group,
+			   struct scsi_iovec *iov, int niov);
+
 EXTERN struct scsi_task *
 iscsi_persistent_reserve_in_sync(struct iscsi_context *iscsi, int lun,
 				 int sa, uint16_t xferlen);
@@ -1162,6 +1464,14 @@ iscsi_report_supported_opcodes_sync(struct iscsi_context *iscsi, int lun,
 				    int opcode, int sa,
 				    uint32_t alloc_len);
 
+EXTERN struct scsi_task *
+iscsi_extended_copy_sync(struct iscsi_context *iscsi, int lun,
+			 struct iscsi_data *param_data);
+
+EXTERN struct scsi_task *
+iscsi_receive_copy_results_sync(struct iscsi_context *iscsi, int lun,
+				int sa, int list_id, int alloc_len);
+
 /*
  * These functions are used when the application wants to specify its own buffers to read the data
  * from the DATA-IN PDUs into, or write the data to DATA-OUT PDUs from.
@@ -1285,6 +1595,16 @@ iscsi_set_tcp_syncnt(struct iscsi_context *iscsi, int value);
 EXTERN void
 iscsi_set_bind_interfaces(struct iscsi_context *iscsi, char * interfaces);
 
+/*
+ * This function is to disable auto reconnect logic.
+ *
+ *  0 - Disable this feature (auto reconnect)
+ *  1 - Enable this feature (no auto reconnect)
+ */
+EXTERN void
+iscsi_set_noautoreconnect(struct iscsi_context *iscsi, int state);
+
+
 /* This function is to set if we should retry a failed reconnect
    
    count is defined as follows:
@@ -1295,6 +1615,12 @@ iscsi_set_bind_interfaces(struct iscsi_context *iscsi, char * interfaces);
 EXTERN void
 iscsi_set_reconnect_max_retries(struct iscsi_context *iscsi, int count);
 
+/* Set to true to have libiscsi use TESTUNITREADY and consume any/all
+   UnitAttentions that may have triggered in the target.
+ */
+EXTERN void
+iscsi_set_no_ua_on_reconnect(struct iscsi_context *iscsi, int state);
+
 #ifdef __cplusplus
 }
 #endif
